__all__ = [
    "ui_style",
]


ui_style = {
    # Standout is caused by man pages that insert a x\b in the output.
    "standout": "bold #44aaff",
    "standout2": "underline #aa8844",
    # Show current search term. (The cursor is not visible).
    "search": "bg:default fg:default reverse",
    "search.current": "bg:#aaff00 #000000 noreverse",
    # UI style
    "statusbar": "reverse",
    "statusbar cursor-position": "noreverse #ffffff bg:#884400",
    "statusbar key": "bold",
    "arg": "reverse #ccaa00",
    "search-toolbar": "bg:#333333 #ffffff",
    "search-toolbar text": "#ffffff",
    "system-toolbar": "bg:#333333 #ffffff",
    "system-toolbar text": "#ffffff",
    "examine": "bg:#333333 #ffffff",
    "examine-text": "bg:#aa88ff #000000",
    "titlebar": "bg:#333333 #ffffff",
    # Messages
    "message": "bg:#bbee88 #222222",
    "loading": "#888888",
    # Help
    "title": "bold #44aaff",
    "subtitle": "bold #8800ff",
    "line": "#888888",
    "keys": "#44aaff",
    "version": "#8800ff",
}



from typing import TYPE_CHECKING

from prompt_toolkit.application import get_app
from prompt_toolkit.filters import Condition, has_focus
from prompt_toolkit.key_binding import KeyBindings
from prompt_toolkit.key_binding.bindings.scroll import (
    scroll_half_page_down,
    scroll_half_page_up,
    scroll_one_line_down,
    scroll_one_line_up,
    scroll_page_down,
    scroll_page_up,
)
from prompt_toolkit.key_binding.key_processor import KeyPressEvent
from prompt_toolkit.keys import Keys
from prompt_toolkit.search import stop_search
from prompt_toolkit.utils import suspend_to_background_supported

if TYPE_CHECKING:
    from .pager import Pager


__all__ = [
    "create_key_bindings",
]

E = KeyPressEvent


def create_key_bindings(pager: "Pager") -> KeyBindings:
    kb = KeyBindings()
    handle = kb.add

    @Condition
    def has_colon() -> bool:
        return pager.in_colon_mode

    @Condition
    def default_focus() -> bool:
        app = get_app()
        return app.layout.current_window == pager.current_source_info.window

    @Condition
    def displaying_help() -> bool:
        return pager.displaying_help

    for c in "01234556789":

        @handle(c, filter=default_focus)
        def _handle_arg(event: E, c: str = c) -> None:
            event.append_to_arg_count(c)

    @handle("q", filter=default_focus, eager=True)
    # Eager because q-any is a registered Vi key binding.
    @handle("Q", filter=default_focus | has_colon)
    @handle("Z", "Z", filter=default_focus)
    def _quit(event: E) -> None:
        "Quit."
        if pager.displaying_help:
            pager.quit_help()
        else:
            event.app.exit()

    @handle(" ", filter=default_focus)
    @handle("f", filter=default_focus)
    @handle("c-f", filter=default_focus)
    @handle("c-v", filter=default_focus)
    def _pagedown(event: E) -> None:
        "Page down."
        scroll_page_down(event)

    @handle("b", filter=default_focus)
    @handle("c-b", filter=default_focus)
    @handle("escape", "v", filter=default_focus)
    def _pageup(event: E) -> None:
        "Page up."
        scroll_page_up(event)

    @handle("d", filter=default_focus)
    @handle("c-d", filter=default_focus)
    def _halfdown(event: E) -> None:
        "Half page down."
        scroll_half_page_down(event)

    @handle("u", filter=default_focus)
    @handle("c-u", filter=default_focus)
    def _halfup(event: E) -> None:
        "Half page up."
        scroll_half_page_up(event)

    @handle("e", filter=default_focus)
    @handle("j", filter=default_focus)
    @handle("c-e", filter=default_focus)
    @handle("c-n", filter=default_focus)
    @handle("c-j", filter=default_focus)
    @handle("c-m", filter=default_focus)
    @handle("down", filter=default_focus)
    def _down(event: E) -> None:
        "Scoll one line down."
        if event.arg > 1:
            # When an argument is given, go this amount of lines down.
            event.current_buffer.auto_down(count=event.arg)
        else:
            scroll_one_line_down(event)

    @handle("y", filter=default_focus)
    @handle("k", filter=default_focus)
    @handle("c-y", filter=default_focus)
    @handle("c-k", filter=default_focus)
    @handle("c-p", filter=default_focus)
    @handle("up", filter=default_focus)
    def _up(event: E) -> None:
        "Scoll one line up."
        if event.arg > 1:
            event.current_buffer.auto_up(count=event.arg)
        else:
            scroll_one_line_up(event)

    @handle(Keys.Escape, "u")
    def _toggle_highlighting(event: E) -> None:
        "Toggle search highlighting."
        pager.highlight_search = not pager.highlight_search

    @handle("=", filter=default_focus)
    @handle(Keys.ControlG, filter=default_focus)
    @handle("f", filter=has_colon)
    def _print_filename(event: E) -> None:
        "Print the current file name."
        pager.message = " {} ".format(pager.current_source.get_name())

    @handle("h", filter=default_focus & ~displaying_help)
    @handle("H", filter=default_focus & ~displaying_help)
    def _help(event: E) -> None:
        "Display Help."
        pager.display_help()

    @handle("g", filter=default_focus)
    @handle("<", filter=default_focus)
    @handle("escape", "<", filter=default_focus)
    def _firstline(event: E) -> None:
        "Go to the first line of the file."
        event.current_buffer.cursor_position = 0

    @handle("G", filter=default_focus)
    @handle(">", filter=default_focus)
    @handle("escape", ">", filter=default_focus)
    def _lastline(event: E) -> None:
        "Go to the last line of the file."
        b = event.current_buffer
        b.cursor_position = len(b.text)

    @handle("m", Keys.Any, filter=default_focus)
    def _mark(event: E) -> None:
        "Mark current position."
        source_info = pager.current_source_info

        source_info.marks[event.data] = (
            event.current_buffer.cursor_position,
            source_info.window.vertical_scroll,
        )

    @handle("'", Keys.Any, filter=default_focus)
    def _goto_mark(event: E) -> None:
        "Go to a previously marked position."
        go_to_mark(event, event.data)

    @handle("c-x", Keys.ControlX, filter=default_focus)
    def _gotomark_dot(event: E) -> None:
        "Same as '."
        go_to_mark(event, ".")

    def go_to_mark(event: E, mark: str) -> None:
        b = event.current_buffer
        source_info = pager.current_source_info
        try:
            if mark == "^":  # Start of file.
                cursor_pos, vertical_scroll = 0, 0
            elif mark == "$":  # End of file - mark.
                cursor_pos, vertical_scroll = len(b.text), 0
            else:  # Custom mark.
                cursor_pos, vertical_scroll = source_info.marks[mark]
        except KeyError:
            pass  # TODO: show warning.
        else:
            b.cursor_position = cursor_pos
            source_info.window.vertical_scroll = vertical_scroll

    @handle("F", filter=default_focus)
    def _follow(event: E) -> None:
        "Forward forever, like 'tail -f'."
        pager.forward_forever = True

    @handle("r", filter=default_focus)
    @handle("R", filter=default_focus)
    def _repaint(event: E) -> None:
        event.app.renderer.clear()

    @Condition
    def search_buffer_is_empty() -> bool:
        "Returns True when the search buffer is empty."
        return pager.search_buffer.text == ""

    @handle(
        "backspace",
        filter=has_focus(pager.search_buffer) & search_buffer_is_empty,
    )
    def _cancel_search(event: E) -> None:
        "Cancel search when backspace is pressed."
        stop_search()

    @Condition
    def line_wrapping_enable() -> bool:
        return pager.current_source_info.wrap_lines

    @handle("left", filter=default_focus & ~line_wrapping_enable)
    @handle("escape", "(", filter=default_focus & ~line_wrapping_enable)
    def _left(event: E) -> None:
        "Scroll half page to the left."
        w = event.app.layout.current_window
        b = event.app.current_buffer

        if w and w.render_info:
            info = w.render_info
            amount = info.window_width // 2

            # Move cursor horizontally.
            value = b.cursor_position - min(
                amount, len(b.document.current_line_before_cursor)
            )
            b.cursor_position = value

            # Scroll.
            w.horizontal_scroll = max(0, w.horizontal_scroll - amount)

    @handle("right", filter=default_focus & ~line_wrapping_enable)
    @handle("escape", ")", filter=default_focus & ~line_wrapping_enable)
    def _right(event: E) -> None:
        "Scroll half page to the right."
        w = event.app.layout.current_window
        b = event.app.current_buffer

        if w and w.render_info:
            info = w.render_info
            amount = info.window_width // 2

            # Move the cursor first to a visible line that is long enough to
            # have the cursor visible after scrolling. (Otherwise, the Window
            # will scroll back.)
            xpos = w.horizontal_scroll + amount

            for line in info.displayed_lines:
                if len(b.document.lines[line]) >= xpos:
                    b.cursor_position = b.document.translate_row_col_to_index(
                        line, xpos
                    )
                    break

            # Scroll.
            w.horizontal_scroll = max(0, w.horizontal_scroll + amount)

    @handle(":", filter=default_focus & ~displaying_help)
    def _colon(event: E) -> None:
        pager.in_colon_mode = True

    @handle("n", filter=has_colon)
    def _next_file(event: E) -> None:
        "Go to next file."
        pager.focus_next_source()

    @handle("p", filter=has_colon)
    def _previous_file(event: E) -> None:
        "Go to previous file."
        pager.focus_previous_source()

    @handle("e", filter=has_colon)
    @handle(Keys.ControlX, Keys.ControlV, filter=default_focus)
    def _examine(event: E) -> None:
        event.app.layout.focus(pager.layout.examine_control)
        pager.in_colon_mode = False

    @handle("d", filter=has_colon)
    def _remove_source(event: E) -> None:
        pager.remove_current_source()

    @handle("backspace", filter=has_colon)
    @handle("q", filter=has_colon, eager=True)
    def _cancel_colon(event: E) -> None:
        pager.in_colon_mode = False

    @handle(Keys.Any, filter=has_colon)
    def _any(event: E) -> None:
        pager.in_colon_mode = False
        pager.message = "No command."

    @handle("c-c", filter=has_focus("EXAMINE"))
    @handle("c-g", filter=has_focus("EXAMINE"))
    def _cancel_examine(event: E) -> None:
        "Cancel 'Examine' input."
        event.app.layout.focus(pager.current_source_info.window)

    @handle("c-z", filter=Condition(lambda: suspend_to_background_supported()))
    def _suspend(event: E) -> None:
        "Suspend to background."
        event.app.suspend_to_background()

    @handle("w")
    def _line_wrapping(event: E) -> None:
        "Enable/disable line wrapping."
        source_info = pager.current_source_info
        source_info.wrap_lines = not source_info.wrap_lines

    return kb



"""
A pager implementation in Python.
"""
__version__ = "3.0.1"

from .pager import Pager
from .source import FormattedTextSource, GeneratorSource, PipeSource, StringSource

__all__ = [
    "FormattedTextSource",
    "GeneratorSource",
    "Pager",
    "PipeSource",
    "StringSource",
]



from typing import TYPE_CHECKING

from prompt_toolkit.filters import Filter

if TYPE_CHECKING:
    from .pager import Pager

__all__ = [
    "HasColon",
    "DisplayingHelp",
]


class _PagerFilter(Filter):
    def __init__(self, pager: "Pager") -> None:
        super().__init__()
        self.pager = pager


class HasColon(_PagerFilter):
    """
    The user typed a ':'.
    """

    def __call__(self) -> bool:
        return self.pager.in_colon_mode


class DisplayingHelp(_PagerFilter):
    def __call__(self) -> bool:
        return self.pager.displaying_help



"""
Pager implementation in Python.
"""
import asyncio
import sys
import threading
import weakref
from typing import Any, Dict, List, Optional, Sequence, Tuple

from prompt_toolkit.application import Application
from prompt_toolkit.buffer import Buffer
from prompt_toolkit.completion import PathCompleter
from prompt_toolkit.document import Document
from prompt_toolkit.enums import EditingMode
from prompt_toolkit.formatted_text import AnyFormattedText, StyleAndTextTuples
from prompt_toolkit.input import Input
from prompt_toolkit.input.defaults import create_input
from prompt_toolkit.layout.layout import Layout
from prompt_toolkit.lexers import Lexer, PygmentsLexer
from prompt_toolkit.output import Output
from prompt_toolkit.styles import Style

from .help import HELP
from .key_bindings import create_key_bindings
from .layout import PagerLayout, create_buffer_window
from .source import DummySource, FileSource, FormattedTextSource, PipeSource, Source
from .style import ui_style

__all__ = [
    "Pager",
    "SourceInfo",
]


class SourceInfo:
    """
    For each opened source, we keep this list of pager data.
    """

    _buffer_counter = 0  # Counter to generate unique buffer names.

    def __init__(self, pager: "Pager", source: Source) -> None:
        self.pager = pager
        self.source = source

        self.buffer = Buffer(read_only=True)

        # List of lines. (Each line is a list of (token, text) tuples itself.)
        self.line_tokens: List[StyleAndTextTuples] = [[]]

        # Marks. (Mapping from mark name to (cursor position, scroll_offset).)
        self.marks: Dict[str, Tuple[int, int]] = {}

        # `Pager` sets this flag when he starts reading the generator of this
        # source in a coroutine.
        self.waiting_for_input_stream = False

        # Enable/disable line wrapping.
        self.wrap_lines = False

        self.window = create_buffer_window(self)


class Pager:
    """
    The Pager main application.

    Usage::
        p = Pager()
        p.add_source(...)
        p.run()

    :param source: :class:`.Source` instance.
    :param lexer: Prompt_toolkit `lexer` instance.
    :param vi_mode: Enable Vi key bindings.
    :param style: Prompt_toolkit `Style` instance.
    :param search_text: `None` or the search string that is highlighted.
    """

    def __init__(
        self,
        *,
        vi_mode: bool = False,
        style: Optional[Style] = None,
        search_text: Optional[str] = None,
        titlebar_tokens: Optional[AnyFormattedText] = None,
        input: Optional[Input] = None,
        output: Optional[Output] = None,
    ) -> None:
        self.sources: List[Source] = []
        self.current_source_index = 0  # Index in `self.sources`.
        self.highlight_search = True
        self.in_colon_mode = False
        self.message: Optional[str] = None
        self.displaying_help = False
        self.search_text = search_text
        self.display_titlebar = bool(titlebar_tokens)
        self.titlebar_tokens = titlebar_tokens or []

        self._dummy_source = DummySource()

        # When this is True, always make sure that the cursor goes to the
        # bottom of the visible content. This is similar to 'tail -f'.
        self.forward_forever = False

        # Status information for all sources. Source -> SourceInfo.
        # (Remember this info as long as the Source object exists.)
        self.source_info: weakref.WeakKeyDictionary[
            Source, SourceInfo
        ] = weakref.WeakKeyDictionary()

        # Create prompt_toolkit stuff.

        def open_file(buff: Buffer) -> bool:
            # Open file.
            self.open_file(buff.text)
            return False

        # Buffer for the 'Examine:' input.
        self.examine_buffer = Buffer(
            name="EXAMINE",
            completer=PathCompleter(expanduser=True),
            accept_handler=open_file,
            multiline=False,
        )

        # Search buffer.
        self.search_buffer = Buffer(multiline=False)

        self.layout = PagerLayout(self)

        # Input/output.
        if input is None:
            # By default, use the stdout device for input.
            # (This makes it possible to pipe data to stdin, but still read key
            # strokes from the TTY).
            input = create_input(sys.stdout)

        bindings = create_key_bindings(self)
        self.application: Application[None] = Application(
            input=input,
            output=output,
            layout=Layout(container=self.layout.container),
            enable_page_navigation_bindings=True,
            key_bindings=bindings,
            style=style or Style.from_dict(ui_style),
            mouse_support=True,
            after_render=self._after_render,
            full_screen=True,
        )

        # Hide message when a key is pressed.
        def key_pressed(_: object) -> None:
            self.message = None

        self.application.key_processor.before_key_press += key_pressed

        if vi_mode:
            self.application.editing_mode = EditingMode.VI

    @classmethod
    def from_pipe(cls, lexer: Optional[Lexer] = None) -> "Pager":
        """
        Create a pager from another process that pipes in our stdin.
        """
        assert not sys.stdin.isatty()
        self = cls()
        self.add_source(
            PipeSource(
                fileno=sys.stdin.fileno(), lexer=lexer, encoding=sys.stdin.encoding
            )
        )
        return self

    @property
    def current_source(self) -> Source:
        "The current `Source`."
        try:
            return self.sources[self.current_source_index]
        except IndexError:
            return self._dummy_source

    @property
    def current_source_info(self) -> SourceInfo:
        try:
            return self.source_info[self.current_source]
        except KeyError:
            return SourceInfo(self, self.current_source)

    def open_file(self, filename: str) -> None:
        """
        Open this file.
        """
        lexer = PygmentsLexer.from_filename(filename, sync_from_start=False)

        try:
            source = FileSource(filename, lexer=lexer)
        except IOError as e:
            self.message = "{}".format(e)
        else:
            self.add_source(source)

    def add_source(self, source: Source) -> SourceInfo:
        """
        Add a new :class:`.Source` instance.
        """
        source_info = SourceInfo(self, source)
        self.source_info[source] = source_info

        self.sources.append(source)

        # Focus
        self.current_source_index = len(self.sources) - 1
        self.application.layout.focus(source_info.window)

        return source_info

    def remove_current_source(self) -> None:
        """
        Remove the current source from the pager.
        (If >1 source is left.)
        """
        if len(self.sources) > 1:
            current_source_index = self.current_source

            # Focus the previous source.
            self.focus_previous_source()

            # Remove the last source.
            self.sources.remove(current_source_index)
        else:
            self.message = "Can't remove the last buffer."

    def focus_previous_source(self) -> None:
        self.current_source_index = (self.current_source_index - 1) % len(self.sources)
        self.application.layout.focus(self.current_source_info.window)
        self.in_colon_mode = False

    def focus_next_source(self) -> None:
        self.current_source_index = (self.current_source_index + 1) % len(self.sources)
        self.application.layout.focus(self.current_source_info.window)
        self.in_colon_mode = False

    def display_help(self) -> None:
        """
        Display help text.
        """
        if not self.displaying_help:
            source = FormattedTextSource(HELP, name="<help>")
            self.add_source(source)
            self.displaying_help = True

    def quit_help(self) -> None:
        """
        Hide the help text.
        """
        if self.displaying_help:
            self.remove_current_source()
            self.displaying_help = False

    def _after_render(self, app: Application[Any]) -> None:
        """
        Each time when the rendering is done, we should see whether we need to
        read more data from the input pipe.
        """
        # When the bottom is visible, read more input.
        # Try at least `info.window_height`, if this amount of data is
        # available.
        info = self.layout.dynamic_body.get_render_info()
        source = self.current_source
        source_info = self.source_info[source]
        b = source_info.buffer
        line_tokens = source_info.line_tokens
        loop = asyncio.get_event_loop()

        if not source_info.waiting_for_input_stream and not source.eof() and info:
            lines_below_bottom = info.ui_content.line_count - info.last_visible_line()

            # Make sure to preload at least 2x the amount of lines on a page.
            if lines_below_bottom < info.window_height * 2 or self.forward_forever:
                # Lines to be loaded.
                lines = [info.window_height * 2 - lines_below_bottom]  # nonlocal

                def handle_content(tokens: StyleAndTextTuples) -> List[str]:
                    """Handle tokens, update `line_tokens`, decrease
                    line count and return list of characters."""
                    data = []
                    for token_char in tokens:
                        char = token_char[1]
                        if char == "\n":
                            line_tokens.append([])

                            # Decrease line count.
                            lines[0] -= 1
                        else:
                            line_tokens[-1].append(token_char)
                        data.append(char)
                    return data

                def insert_text(list_of_fragments: Sequence[str]) -> None:
                    document = Document(
                        b.text + "".join(list_of_fragments), b.cursor_position
                    )
                    b.set_document(document, bypass_readonly=True)

                    if self.forward_forever:
                        b.cursor_position = len(b.text)

                    # Schedule redraw.
                    self.application.invalidate()

                    source_info.waiting_for_input_stream = False

                def receive_content_from_generator() -> None:
                    "(in executor) Read data from generator."
                    # Call `read_chunk` as long as we need more lines.
                    while lines[0] > 0 and not source.eof():
                        tokens = source.read_chunk()
                        data = handle_content(tokens)
                        loop.call_soon(insert_text, data)

                # Set 'waiting_for_input_stream' and render.
                source_info.waiting_for_input_stream = True
                self.application.invalidate()

                # Execute receive_content_from_generator in thread.
                # (Don't use 'run_in_executor', because we need a daemon.
                t = threading.Thread(target=receive_content_from_generator)
                t.daemon = True
                t.start()

    def _before_run(self) -> None:
        # Set search highlighting.
        if self.search_text:
            self.application.current_search_state.text = self.search_text

    def run(self) -> None:
        """
        Create an event loop for the application and run it.
        """
        try:
            self._before_run()
            return self.application.run()
        finally:
            # XXX: Close all sources which are opened by the pager itself.
            pass

    async def run_async(self) -> None:
        """
        Create an event loop for the application and run it.
        """
        try:
            self._before_run()
            return await self.application.run_async()
        finally:
            # XXX: Close all sources which are opened by the pager itself.
            pass



import weakref
from typing import TYPE_CHECKING, Optional

from prompt_toolkit.application import get_app
from prompt_toolkit.enums import SYSTEM_BUFFER
from prompt_toolkit.filters import Condition, HasArg, HasSearch, has_focus
from prompt_toolkit.formatted_text import HTML, AnyFormattedText, StyleAndTextTuples
from prompt_toolkit.layout.containers import (
    ConditionalContainer,
    Container,
    Float,
    FloatContainer,
    HSplit,
    VSplit,
    Window,
    WindowAlign,
    WindowRenderInfo,
)
from prompt_toolkit.layout.controls import BufferControl, FormattedTextControl
from prompt_toolkit.layout.menus import MultiColumnCompletionsMenu
from prompt_toolkit.layout.processors import (
    BeforeInput,
    ConditionalProcessor,
    HighlightIncrementalSearchProcessor,
    HighlightMatchingBracketProcessor,
    HighlightSearchProcessor,
    HighlightSelectionProcessor,
    Processor,
    TabsProcessor,
    Transformation,
    TransformationInput,
)
from prompt_toolkit.lexers import SimpleLexer
from prompt_toolkit.widgets.toolbars import (
    FormattedTextToolbar,
    SearchToolbar,
    SystemToolbar,
)

from .filters import HasColon

if TYPE_CHECKING:
    from .pager import Pager, SourceInfo

__all__ = [
    "PagerLayout",
]


class _EscapeProcessor(Processor):
    """
    Interpret escape sequences like less/more/most do.
    """

    def __init__(self, source_info: "SourceInfo") -> None:
        self.source_info = source_info

    def apply_transformation(self, ti: TransformationInput) -> Transformation:
        tokens = self.source_info.line_tokens[ti.lineno]
        return Transformation(tokens[:])


class _Arg(ConditionalContainer):
    def __init__(self) -> None:
        def get_text() -> str:
            app = get_app()
            if app.key_processor.arg is not None:
                return " %s " % app.key_processor.arg
            else:
                return ""

        super().__init__(
            Window(
                FormattedTextControl(get_text),
                style="class:arg",
                align=WindowAlign.RIGHT,
            ),
            filter=HasArg(),
        )


class Titlebar(FormattedTextToolbar):
    """
    Displayed at the top.
    """

    def __init__(self, pager: "Pager") -> None:
        def get_tokens() -> AnyFormattedText:
            return pager.titlebar_tokens

        super().__init__(get_tokens)


class MessageToolbarBar(FormattedTextToolbar):
    """
    Pop-up (at the bottom) for showing error/status messages.
    """

    def __init__(self, pager: "Pager") -> None:
        def get_tokens():
            return [("class:message", pager.message)] if pager.message else []

        super().__init__(get_tokens)


class _DynamicBody(Container):
    def __init__(self, pager: "Pager") -> None:
        self.pager = pager
        self._bodies: weakref.WeakKeyDictionary[
            str, Window
        ] = weakref.WeakKeyDictionary()  # Map buffer_name to Window.

    def get_buffer_window(self) -> Window:
        "Return the Container object according to which Buffer/Source is visible."
        return self.pager.current_source_info.window

    def reset(self) -> None:
        for body in self._bodies.values():
            body.reset()

    def get_render_info(self) -> Optional[WindowRenderInfo]:
        return self.get_buffer_window().render_info

    def preferred_width(self, *a, **kw):
        return self.get_buffer_window().preferred_width(*a, **kw)

    def preferred_height(self, *a, **kw):
        return self.get_buffer_window().preferred_height(*a, **kw)

    def write_to_screen(self, *a, **kw):
        return self.get_buffer_window().write_to_screen(*a, **kw)

    def get_children(self):
        return [self.get_buffer_window()]


class PagerLayout:
    def __init__(self, pager: "Pager") -> None:
        self.pager = pager
        self.dynamic_body = _DynamicBody(pager)

        # Build an interface.
        has_colon = HasColon(pager)

        self.examine_control: BufferControl = BufferControl(
            buffer=pager.examine_buffer,
            lexer=SimpleLexer(style="class:examine,examine-text"),
            input_processors=[BeforeInput(lambda: [("class:examine", " Examine: ")])],
        )

        self.search_toolbar = SearchToolbar(
            vi_mode=True, search_buffer=pager.search_buffer
        )

        self.container = FloatContainer(
            content=HSplit(
                [
                    ConditionalContainer(
                        content=Titlebar(pager),
                        filter=Condition(lambda: pager.display_titlebar),
                    ),
                    self.dynamic_body,
                    self.search_toolbar,
                    SystemToolbar(),
                    ConditionalContainer(
                        content=VSplit(
                            [
                                Window(
                                    height=1,
                                    content=FormattedTextControl(
                                        self._get_statusbar_left_tokens
                                    ),
                                    style="class:statusbar",
                                ),
                                Window(
                                    height=1,
                                    content=FormattedTextControl(
                                        self._get_statusbar_right_tokens
                                    ),
                                    style="class:statusbar.cursorposition",
                                    align=WindowAlign.RIGHT,
                                ),
                            ]
                        ),
                        filter=~HasSearch()
                        & ~has_focus(SYSTEM_BUFFER)
                        & ~has_colon
                        & ~has_focus("EXAMINE"),
                    ),
                    ConditionalContainer(
                        content=Window(
                            FormattedTextControl(" :"), height=1, style="class:examine"
                        ),
                        filter=has_colon,
                    ),
                    ConditionalContainer(
                        content=Window(
                            self.examine_control, height=1, style="class:examine"
                        ),
                        filter=has_focus(pager.examine_buffer),
                    ),
                ]
            ),
            floats=[
                Float(right=0, height=1, bottom=1, content=_Arg()),
                Float(
                    bottom=1,
                    left=0,
                    right=0,
                    height=1,
                    content=ConditionalContainer(
                        content=MessageToolbarBar(pager),
                        filter=Condition(lambda: bool(pager.message)),
                    ),
                ),
                Float(
                    right=0,
                    height=1,
                    bottom=1,
                    content=ConditionalContainer(
                        content=FormattedTextToolbar(
                            lambda: [("class:loading", " Loading... ")],
                        ),
                        filter=Condition(
                            lambda: pager.current_source_info.waiting_for_input_stream
                        ),
                    ),
                ),
                Float(xcursor=True, ycursor=True, content=MultiColumnCompletionsMenu()),
            ],
        )

    def _get_statusbar_left_tokens(self) -> HTML:
        """
        Displayed at the bottom left.
        """
        if self.pager.displaying_help:
            return HTML(" HELP -- Press <key>[q]</key> when done")
        else:
            return HTML(" (press <key>[h]</key> for help or <key>[q]</key> to quit)")

    def _get_statusbar_right_tokens(self) -> StyleAndTextTuples:
        """
        Displayed at the bottom right.
        """
        source_info = self.pager.source_info[self.pager.current_source]
        buffer = source_info.buffer
        document = buffer.document
        row = document.cursor_position_row + 1
        col = str(document.cursor_position_col + 1)

        if source_info.wrap_lines:
            col = "WRAP"

        if self.pager.current_source.eof():
            percentage = int(100 * row / document.line_count)
            return [
                (
                    "class:statusbar,cursor-position",
                    " (%s,%s) %s%% " % (row, col, percentage),
                )
            ]
        else:
            return [("class:statusbar,cursor-position", " (%s,%s) " % (row, col))]


def create_buffer_window(source_info: "SourceInfo") -> Window:
    """
    Window for the main content.
    """
    pager = source_info.pager

    input_processors = [
        ConditionalProcessor(
            processor=_EscapeProcessor(source_info),
            filter=Condition(lambda: not bool(source_info.source.lexer)),
        ),
        TabsProcessor(),
        ConditionalProcessor(
            processor=HighlightSearchProcessor(),
            filter=Condition(lambda: pager.highlight_search),
        ),
        ConditionalProcessor(
            processor=HighlightIncrementalSearchProcessor(),
            filter=Condition(lambda: pager.highlight_search),
        ),
        HighlightSelectionProcessor(),
        HighlightMatchingBracketProcessor(),
    ]

    @Condition
    def wrap_lines() -> bool:
        return source_info.wrap_lines

    return Window(
        always_hide_cursor=True,
        wrap_lines=wrap_lines,
        content=BufferControl(
            buffer=source_info.buffer,
            lexer=source_info.source.lexer,
            input_processors=input_processors,
            include_default_input_processors=False,
            preview_search=True,
            search_buffer_control=pager.layout.search_toolbar.control,
        ),
    )



#!/usr/bin/env python
"""
pypager: A pure Python pager application.
"""
import argparse
import os
import sys

from prompt_toolkit.lexers import PygmentsLexer

from pypager.pager import Pager
from pypager.source import FileSource

__all__ = [
    "run",
]


def run() -> None:
    if not sys.stdin.isatty():
        pager = Pager.from_pipe()
        pager.run()
    else:
        parser = argparse.ArgumentParser(description="Browse through a text file.")
        parser.add_argument(
            "filename", metavar="filename", nargs="+", help="The file to be displayed."
        )
        parser.add_argument("--vi", help="Prefer Vi key bindings.", action="store_true")
        parser.add_argument(
            "--emacs", help="Prefer Emacs key bindings.", action="store_true"
        )

        args = parser.parse_args()

        # Determine input mode.
        vi_mode = "vi" in os.environ.get("EDITOR", "").lower()
        if args.vi:
            vi_mode = True
        if args.emacs:
            vi_mode = False

        pager = Pager(vi_mode=vi_mode)

        # Open files.
        for filename in args.filename:
            # When a filename is given, take a lexer from that filename.
            lexer = PygmentsLexer.from_filename(filename, sync_from_start=False)

            pager.add_source(FileSource(filename, lexer=lexer))

        # Run UI.
        pager.run()



import sys

import prompt_toolkit
from prompt_toolkit.formatted_text import HTML

import pypager

python_version = sys.version_info
ptk_version = prompt_toolkit.__version__
pypager_version = pypager.__version__


HELP = (
    HTML(
        """
            <title>SUMMARY OF COMMANDS</title>

 <keys> h  H             </keys> Display this help.
 <keys> q  Q  ZZ         </keys> Exit.

 <line>------------------------------------------------------</line>

  <subtitle> Moving </subtitle>

 <keys>e  ^E  j  ^N  CR  </keys> Forward one line.
 <keys>y  ^Y  k  ^K  ^P  </keys> Backward one line.
 <keys>f  ^F  ^V  SPACE  </keys> Forward one window.
 <keys>b  ^B  ESC-v      </keys> Backward one window.
 <keys>d  ^D             </keys> Forward one half-window.
 <keys>u  ^U             </keys> Backward one half-window.
 <keys>ESC-)  RightArrow </keys> Left one half screen width.
 <keys>ESC-(  LeftArrow  </keys> Right one half screen width.
 <keys>F                 </keys> Forward forever; like "tail -f"
 <keys>r  R  ^R  ^L      </keys> Repaint screen.

  <subtitle> SEARCHING </subtitle>

 <keys>/pattern          </keys> Search forward.
 <keys>?pattern          </keys> Search backward.
 <keys>n                 </keys> Repeat previous search.
 <keys>N                 </keys> Repeat previous search in reverse direction.
 <keys>ESC-u             </keys> Undo (toggle) search highlighting.

  <subtitle> JUMPING </subtitle>

 <keys> g  &lt;  ESC-&lt;</keys>       Go to the first line in file.
 <keys> G  &gt;  ESC-&gt;</keys>       Go to the last line in file.

 <keys>m&lt;letter&gt;   </keys>       Mark the current position with &lt;letter&gt;
 <keys>'&lt;letter&gt;   </keys>       Go to a previously marked position.
 <keys>^X^X              </keys> Same as <keys>'.</keys>

    A mark is any upper-case or lower-case letter.
    Certain marks are predefined.
        <keys>^</keys>  means  beginning of the file
        <keys>$</keys>  means  end of the file

  <subtitle> CHANGING FILES </subtitle>

  <keys>:e               </keys> Examine a new file.
  <keys>^E^V             </keys> Same as :e.
  <keys>:n               </keys> Examine the next file from the command line.
  <keys>:p               </keys> Examine the previous file from the command line.
  <keys>:d               </keys> Delete the current file from the command line list.
  <keys>=  ^G  :f        </keys> Print current file name.

  <subtitle> Extras </subtitle>

  <keys>w                </keys> Enable/disable <b>line wrapping</b>.

  <subtitle> About Pypager </subtitle>

  Pypager is a <u>prompt_toolkit</u> application.

  - Pypager version:        <version>%s</version>
  - Python version:         <version>%s.%s.%s</version>
  - prompt_toolkit version: <version>%s</version>

"""
    )
    % (
        pypager_version,
        python_version[0],
        python_version[1],
        python_version[2],
        ptk_version,
    )
)



"""
Input source for a pager.
(pipe or generator.)
"""
import os
from abc import ABCMeta, abstractmethod
from codecs import getincrementaldecoder
from typing import Dict, Generator, Optional, Sequence, Union

from prompt_toolkit.formatted_text import (
    AnyFormattedText,
    StyleAndTextTuples,
    to_formatted_text,
)
from prompt_toolkit.layout.utils import explode_text_fragments
from prompt_toolkit.lexers import Lexer
from prompt_toolkit.output.vt100 import BG_ANSI_COLORS, FG_ANSI_COLORS
from prompt_toolkit.output.vt100 import _256_colors as _256_colors_table
from prompt_toolkit.styles import Attrs

__all__ = [
    "Source",
    "DummySource",
    "PipeSource",
    "GeneratorSource",
    "StringSource",
    "FormattedTextSource",
]


class Source(metaclass=ABCMeta):
    #: The lexer to be used in the layout.
    lexer: Optional[Lexer] = None

    @abstractmethod
    def get_name(self) -> str:
        "Return the filename or name for this input."

    @abstractmethod
    def eof(self) -> bool:
        "Return True when we reached the end of the input."

    @abstractmethod
    def read_chunk(self) -> StyleAndTextTuples:
        """
        Read data from input. Return a list of token/text tuples.

        This can be blocking and will be called in another thread.
        """

    def close(self) -> None:
        pass


class DummySource(Source):
    """
    Empty source.
    """

    def get_name(self) -> str:
        return ""

    def eof(self) -> bool:
        return True

    def read_chunk(self) -> StyleAndTextTuples:
        return []


class PipeSource(Source):
    """
    When input is read from another process that is chained to use through a
    unix pipe.
    """

    def __init__(
        self,
        fileno: int,
        lexer: Optional[Lexer] = None,
        name: str = "<stdin>",
        encoding: str = "utf-8",
    ) -> None:
        self.fileno = fileno
        self.lexer = lexer
        self.name = name

        self._line_tokens: StyleAndTextTuples = []
        self._eof = False

        # Default style attributes.
        self._attrs = Attrs(
            color=None,
            bgcolor=None,
            bold=False,
            underline=False,
            strike=False,
            italic=False,
            blink=False,
            reverse=False,
            hidden=False,
        )

        # Start input parser.
        self._parser = self._parse_corot()
        next(self._parser)

        # Create incremental decoder for decoding stdin.
        # We can not just do `os.read(stdin.fileno(), 1024).decode('utf-8')`,
        # because it could be that we are in the middle of a utf-8 byte
        # sequence.
        self._stdin_decoder_cls = getincrementaldecoder(encoding)
        self._stdin_decoder = self._stdin_decoder_cls(errors="ignore")

    def get_name(self) -> str:
        return self.name

    def eof(self) -> bool:
        return self._eof

    def _get_data(self) -> str:
        data = os.read(self.fileno, 1024)

        # Nothing more to read, stream is closed.
        if data == b"":
            self._eof = True
            return ""

        return self._stdin_decoder.decode(data)

    def read_chunk(self) -> StyleAndTextTuples:
        # Content is ready for reading on stdin.
        data = self._get_data()

        # Send input data to the parser.
        for c in data:
            self._parser.send(c)

        # Return the tokens from the parser.
        # (Don't return the last token yet, because the parser should
        # be able to pop if the input starts with \b).
        if self._eof:
            tokens = self._line_tokens[:]
            del self._line_tokens[:]
        else:
            tokens = self._line_tokens[:-1]
            del self._line_tokens[:-1]

        return tokens

    def _parse_corot(self) -> Generator[None, str, None]:
        """
        Coroutine that parses the pager input.
        A \b with any character before should make the next character standout.
        A \b with an underscore before should make the next character emphasized.
        """
        backspace_style = ""  # Style created by backspace characters.
        line_tokens = self._line_tokens
        replace_one_token = False

        while True:
            csi = False
            c = yield

            if c == "\b":
                # Handle \b escape codes from man pages.
                if line_tokens:
                    last_char = line_tokens[-1][1]
                    line_tokens.pop()
                    replace_one_token = True
                    if last_char == "_":
                        backspace_style = "class:standout2"
                    else:
                        backspace_style = "class:standout"
                continue

            elif c == "\x1b":
                # Start of color escape sequence.
                square_bracket = yield
                if square_bracket == "[":
                    csi = True
                else:
                    continue
            elif c == "\x9b":
                csi = True

            if csi:
                # Got a CSI sequence. Color codes are following.
                current = ""
                params = []
                while True:
                    char = yield
                    if char.isdigit():
                        current += char
                    else:
                        params.append(min(int(current or 0), 9999))
                        if char == ";":
                            current = ""
                        elif char == "m":
                            # Set attributes and token.
                            self._select_graphic_rendition(
                                params
                            )  ### TODO: use inline style.
                            #### token = ('C', ) + self._attrs
                            break
                        else:
                            # Ignore unspported sequence.
                            break
            else:
                line_tokens.append((self._get_attrs_style() + " " + backspace_style, c))
                if replace_one_token:
                    backspace_style = ""

    def _select_graphic_rendition(self, attrs: Sequence[int]) -> None:
        """
        Taken a list of graphics attributes and apply changes to Attrs.
        """
        # NOTE: This function is almost literally taken from Pymux.
        #       if something is wrong, please report there as well!
        #       https://github.com/jonathanslenders/pymux
        replace: Dict[str, Union[bool, str, None]] = {}

        if not attrs:
            attrs = [0]
        else:
            attrs = list(attrs[::-1])

        while attrs:
            attr = attrs.pop()

            if attr in _fg_colors:
                replace["color"] = _fg_colors[attr]
            elif attr in _bg_colors:
                replace["bgcolor"] = _bg_colors[attr]
            elif attr == 1:
                replace["bold"] = True
            elif attr == 3:
                replace["italic"] = True
            elif attr == 4:
                replace["underline"] = True
            elif attr == 5:
                replace["blink"] = True
            elif attr == 6:
                replace["blink"] = True  # Fast blink.
            elif attr == 7:
                replace["reverse"] = True
            elif attr == 22:
                replace["bold"] = False
            elif attr == 23:
                replace["italic"] = False
            elif attr == 24:
                replace["underline"] = False
            elif attr == 25:
                replace["blink"] = False
            elif attr == 27:
                replace["reverse"] = False
            elif not attr:
                replace = {}
                self._attrs = Attrs(
                    color=None,
                    bgcolor=None,
                    bold=False,
                    underline=False,
                    strike=False,
                    italic=False,
                    blink=False,
                    reverse=False,
                    hidden=False,
                )

            elif attr in (38, 48):
                n = attrs.pop()

                # 256 colors.
                if n == 5:
                    if attr == 38:
                        m = attrs.pop()
                        replace["color"] = _256_colors.get(1024 + m)
                    elif attr == 48:
                        m = attrs.pop()
                        replace["bgcolor"] = _256_colors.get(1024 + m)

                # True colors.
                if n == 2:
                    try:
                        color_str = "%02x%02x%02x" % (
                            attrs.pop(),
                            attrs.pop(),
                            attrs.pop(),
                        )
                    except IndexError:
                        pass
                    else:
                        if attr == 38:
                            replace["color"] = color_str
                        elif attr == 48:
                            replace["bgcolor"] = color_str

        self._attrs = self._attrs._replace(**replace)  # type: ignore

    def _get_attrs_style(self) -> str:
        result = []
        attrs = self._attrs

        if attrs.color:
            result.append(" fg:{} ".format(attrs.color))
        if attrs.bgcolor:
            result.append(" bg:{} ".format(attrs.bgcolor))
        if attrs.bold:
            result.append(" bold ")
        if attrs.italic:
            result.append(" italic ")
        if attrs.underline:
            result.append(" underline ")
        if attrs.blink:
            result.append(" blink ")
        if attrs.reverse:
            result.append(" reverse ")

        # Recent versions of Groff (used for man pages) use bold and underline
        # escape sequences rather then backslash-style escape sequences. Apply
        # the standout/standout2 styles anyway so that we get colored output.
        # This way, people don't have to set GROFF_NO_SGR=1.
        if attrs.bold and not attrs.color:
            result.append(" class:standout ")
        if attrs.underline and not attrs.color:
            result.append(" class:standout2 ")

        return "".join(result)


class FileSource(PipeSource):
    def __init__(self, filename: str, lexer: Optional[Lexer] = None) -> None:
        self.fp = open(filename, "rb")

        super().__init__(self.fp.fileno(), lexer=lexer, name=filename)

    def close(self) -> None:
        self.fp.close()


# Mapping of the ANSI color codes to their names.
_fg_colors = {v: k for k, v in FG_ANSI_COLORS.items()}
_bg_colors = {v: k for k, v in BG_ANSI_COLORS.items()}

# Mapping of the escape codes for 256colors to their 'ffffff' value.
_256_colors = {}

for i, (r, g, b) in enumerate(_256_colors_table.colors):
    _256_colors[1024 + i] = "%02x%02x%02x" % (r, g, b)


class GeneratorSource(Source):
    """
    When the input is coming from a Python generator.
    """

    def __init__(
        self,
        generator: Generator[StyleAndTextTuples, None, None],
        lexer: Optional[Lexer] = None,
        name: str = "",
    ) -> None:
        self._eof = False
        self.generator = generator
        self.lexer = lexer
        self.name = name

    def get_name(self) -> str:
        return self.name

    def eof(self) -> bool:
        return self._eof

    def read_chunk(self) -> StyleAndTextTuples:
        "Read data from input. Return a list of token/text tuples."
        try:
            return explode_text_fragments(next(self.generator))
        except StopIteration:
            self._eof = True
            return []


class StringSource(Source):
    """
    Take a Python string is input for the pager.
    """

    def __init__(
        self, text: str, lexer: Optional[Lexer] = None, name: str = ""
    ) -> None:
        self.text = text
        self.lexer = lexer
        self.name = name
        self._read = False

    def get_name(self) -> str:
        return self.name

    def eof(self) -> bool:
        return self._read

    def read_chunk(self) -> StyleAndTextTuples:
        if self._read:
            return []
        else:
            self._read = True
            return explode_text_fragments([("", self.text)])


class FormattedTextSource(Source):
    """
    Take any kind of prompt_toolkit formatted text as input for the pager.
    """

    def __init__(self, formatted_text: AnyFormattedText, name: str = "") -> None:
        self.formatted_text = to_formatted_text(formatted_text)
        self.name = name
        self._read = False

    def get_name(self) -> str:
        return self.name

    def eof(self) -> bool:
        return self._read

    def read_chunk(self) -> StyleAndTextTuples:
        if self._read:
            return []
        else:
            self._read = True
            return explode_text_fragments(self.formatted_text)



#!/usr/bin/env python
import os

from setuptools import find_packages, setup

import pypager

with open(os.path.join(os.path.dirname(__file__), "README.rst")) as f:
    long_description = f.read()

install_requires = [
    "prompt_toolkit>=3.0.0,<3.1.0",
    "pygments",
]

setup(
    name="pypager",
    author="Jonathan Slenders",
    version=pypager.__version__,
    license="LICENSE",
    url="https://github.com/jonathanslenders/pypager",
    description='Pure Python pager (like "more" and "less").',
    long_description=long_description,
    packages=find_packages("."),
    python_requires=">=3.6",
    install_requires=install_requires,
    entry_points={
        "console_scripts": [
            "pypager = pypager.entry_points.run_pypager:run",
        ]
    },
)
